/*
 * dproc.c - Pyramid DC/OSx and Reliant UNIX process access functions for lsof
 */


/*
 * Copyright 1996 Purdue Research Foundation, West Lafayette, Indiana
 * 47907.  All rights reserved.
 *
 * Written by Victor A. Abell
 *
 * This software is not subject to any license of the American Telephone
 * and Telegraph Company or the Regents of the University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor Purdue University are responsible for any
 *    consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors and Purdue
 *    University must appear in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This notice may not be removed or altered.
 */

#ifndef lint
static char copyright[] =
"@(#) Copyright 1996 Purdue Research Foundation.\nAll rights reserved.\n";
static char *rcsid = "$Id: dproc.c,v 1.1 96/12/09 07:14:47 abe Exp $";
#endif

#include "lsof.h"

#if     defined(HASNCACHE)
#include <sys/dnlc.h>
#endif  /* defined(HASNCACHE) */


/*
 * Local function prototypes.
 */

_PROTOTYPE(static void get_kernel_access,(void));

#if     defined(HASNCACHE)
_PROTOTYPE(static void ncache_load,(void));
#endif  /* defined(HASNCACHE) */

_PROTOTYPE(static void readfsinfo,(void));
_PROTOTYPE(static void process_text,(struct as *pa));


/*
 * gather_proc_info() -- gather process information
 */

void
gather_proc_info()
{
	struct dirent *dp;
	struct file *f;
	int i, j, pgrp, pid;
	MALLOC_S len;
	static struct proc *p = (struct proc *)NULL;
	int pfd = -1;
	static DIR *pfsfd = (DIR *)NULL;
	struct pid pids;
	static char *pn = (char *)NULL;
	static MALLOC_S pna = 0;
	static char *pnp;
	static MALLOC_S pnpl;
	static MALLOC_S pnsp;
	short pss, sf;
	static struct user *u = (struct user *)NULL;
	uid_t uid;
/*
 * Allocate proc structure buffer.
 */
	if (!p) {
	    if (!(p = (struct proc *)malloc(sizeof(struct proc)))) {
		(void) fprintf(stderr,
		    "%s: can't allocate %d bytes for proc structure\n",
		    Pn, sizeof(struct proc));
		Exit(1);
	    }
	}
/*
 * Allocate a properly aligned user structure buffer.
 */
	if (!u) {
	    i = sysconf(_SC_PAGESIZE);
	    if (!(u = (struct user *)memalign(i, (sizeof(struct user) + i)))) {
		(void) fprintf(stderr,
		    "%s: can't memalign user structure buffer\n", Pn);
		Exit(1);
	    }
	}
/*
 * Prepare the path name buffer for reading individual /proc/<number> files.
 */
	if (!pn) {
	    pnpl = strlen(PROCFS) + 1;
	    pna = (MALLOC_S)(pnpl + 1 + 10);
	    if ((pn = (char *)malloc(pna)) == (char *)NULL) {
		(void) fprintf(stderr,
		    "%s: can't allocate %d bytes for %s name buffer\n",
		    Pn, pna, PROCFS);
		Exit(1);
	    }
	    (void) sprintf(pn, "%s/", PROCFS);
	    pnp = &pn[pnpl];
	    pnsp = (MALLOC_S)(pna - pnpl - 1);
	}

#if     defined(HASNCACHE)
/*
 * Read the kernel name cache.
 */
	ncache_load();
#endif  /* defined(HASNCACHE) */

/*
 * Open access to the /proc directory, or position it to its beginning.
 */
	if (!pfsfd) {
	    if (!(pfsfd = opendir(PROCFS))) {
		(void) fprintf(stderr, "%s: can't open %s\n", Pn, PROCFS);
		Exit(1);
	    }
	} else
	    (void) rewinddir(pfsfd);
/*
 * Examine proc structures and their associated information.
 */
	while ((dp = readdir(pfsfd))) {

	/*
	 * Process next procfs directory entry.
	 */
	    if (pfd >= 0) {
		(void) close(pfd);
		pfd = -1;
	    }
	    for (i = len = 0; dp->d_name[i]; i++) {
		if (dp->d_name[i] < '0' || dp->d_name[i] > '9')
		    break;
		len++;
	    }
	    if (dp->d_name[i] || !len)
		continue;
	    if (len > pnsp) {

	    /*
	     * Expand the directoru entry buffer.
	     */
		pna = (MALLOC_S)(pnpl + len + 10 + 1);
		if (!(pn = (char *)realloc(pn, pna))) {
		    (void) fprintf(stderr,
			"%s: can't realloc %d bytes for %s path name buffer\n",
			Pn, pna, PROCFS);
		    Exit(1);
		}
		pnp = &pn[pnpl];
		pnsp = (MALLOC_S)(pna - pnpl - 1);
	    }
	    (void) strcpy(pnp, dp->d_name);
	    if ((pfd = open(pn, O_RDONLY, 0)) < 0)
		continue;
	    if (ioctl(pfd, PIOCGETPR, p) < 0)
		continue;
	    if (p->p_stat == 0 || p->p_stat == SZOMB)
		continue;
	/*
	 * Get Process ID, Process group ID, and User ID.
	 */
	    if (!p->p_pidp
	    ||  kread((KA_T)p->p_pidp, (char *)&pids, sizeof(pids)))
		continue;
	    pid = (int)pids.pid_id;
	    if (!p->p_pgidp
	    ||  kread((KA_T)p->p_pgidp, (char *)&pids, sizeof(pids)))
		continue;
	    pgrp = (int)pids.pid_id;
	    uid = (uid_t)p->p_uid;
	    if (is_proc_excl(pid, pgrp, (UID_ARG)uid, &pss, &sf))
		continue;
	/*
	 * Get the user area associated with the process.
	 */
	    if (ioctl(pfd, PIOCGETU, u) < 0)
		continue;
	/*
	 * Allocate a local process structure.
	 */
	    if (is_cmd_excl(p->p_comm, &pss, &sf))
		continue;
	    alloc_lproc(pid, pgrp, (UID_ARG)uid, p->p_comm, (int)pss, (int)sf);
	    Plf = NULL;
	/*
	 * Save current working directory information.
	 */
	    if (u->u_cdir) {
		alloc_lfile(CWD, -1);
		process_node((caddr_t)u->u_cdir);
		if (Lf->sf)
		    link_lfile();
	    }
	/*
	 * Save root directory information.
	 */
	    if (u->u_rdir) {
		alloc_lfile(RTD, -1);
		process_node((caddr_t)u->u_rdir);
		if (Lf->sf)
		    link_lfile();
	    }
	/*
	 * Print information on the text file.
	 */
	    if (p->p_as && Nl[X_SEGVN].n_value)
		process_text(p->p_as);
	/*
	 * Save information on file descriptors.
	 */
	    for (i = 0, j = 0; i < u->u_nofiles; i++) {
		if (++j > NFPCHUNK) {
		    if (!u->u_flist.uf_next)
			break;
		    if (kread((KA_T)u->u_flist.uf_next, (char *)&u->u_flist,
		    	      sizeof(struct ufchunk)))
			break;
		    j = 1;
		}
		f = u->u_flist.uf_ofile[j-1];
		if (f) {
		    alloc_lfile(NULL, i);
		    process_file(f);
		    if (Lf->sf)
			link_lfile();
		}
	    }
	/*
	 * Examine results.
	 */
	    if (examine_lproc())
		break;
	}
/*
 * Close directory accesses before returning.
 */
	if (pfd >= 0) {
	    (void) close(pfd);
	    pfd = -1;
	}
	if (!RptTm && pfsfd) {
	    (void) closedir(pfsfd);
	    pfsfd = (DIR *)NULL;
	}
}


/*
 * get_kernel_access() - get access to kernel memory
 */

static void
get_kernel_access()
{

#if	defined(WILLDROPGID)
/*
 * If kernel memory isn't coming from KMEM, drop setgid permission
 * before attempting to open the (Memory) file.
 */
	if (Memory)
	    (void) dropgid();
#else	/* !defined(WILLDROPGID) */
/*
 * See if the non-KMEM memory file is readable.
 */
	if (Memory && !is_readable(Memory, 1))
	    Exit(1);
#endif	/* defined(WILLDROPGID) */

/*
 * Open kernel memory access.
 */
	if ((Kd = open(Memory ? Memory : KMEM, O_RDONLY, 0)) < 0) {
	    (void) fprintf(stderr, "%s: can't open %s: %s\n", Pn,
		Memory ? Memory : KMEM, strerror(errno));
	    Exit(1);
	}

#if	defined(WILLDROPGID)
/*
 * Drop setgid permission, if necessary.
 */
	if (!Memory)
	    (void) dropgid();
#else	/* !defined(WILLDROPGID) */
/*
 * See if the name list file is readable.
 */
	if (Nmlst && !is_readable(Nmlst, 1))
	    Exit(1);
#endif	/* defined(WILLDROPGID) */

/*
 * Access kernel symbols and values.
 */
        if (nlist(Nmlst ? Nmlst : N_UNIX, Nl) < 0) {
	    (void) fprintf(stderr, "%s: can't read kernel name list from %s\n",
		Pn, Nmlst ? Nmlst : N_UNIX);
	    Exit(1);
	}
	if (Nl[X_CLONMAJ].n_value
	&&  kread((KA_T)Nl[X_CLONMAJ].n_value, (char *)&CloneMaj,
		  sizeof(CloneMaj)) == 0)
	{
	    HaveCloneMaj = 1;
	}
}


/*
 * initialize() - perform all initialization
 */

void
initialize()
{
	get_kernel_access();
	readfsinfo();
}


/*
 * kread() - read from kernel memory
 */

int
kread(addr, buf, len)
	KA_T addr;			/* kernel memory address */
	char *buf;			/* buffer to receive data */
	READLEN_T len;			/* length to read */
{
	int br;

	if (lseek(Kd, (long)addr, L_SET) == (long)-1L)
		return(-1);
	br = read(Kd, buf, len);
	return((br == len) ? 0 : 1);
}


/*
 * process_text() - process text access information
 */

static void
process_text(pa)
	struct as *pa;			/* address space description pointer */
{
	struct as as;
	int i, j, k, l;
	struct seg s;
	union segdata {
		struct segvn_data vn;
		struct segdev_data dev;
		struct segmap_data map;
	} sd;
	struct vnode *v[MAXSEGS];
/*
 * Get address space description.
 */
	if (kread((KA_T)pa, (char *)&as, sizeof(as))) {
		alloc_lfile(" txt", -1);
		(void) sprintf(Namech, "can't read text segment list (%#x)",
			pa);
		enter_nm(Namech);
		if (Lf->sf)
			link_lfile();
		return;
	}
/*
 * Loop through the segments.  The loop should stop when the segment
 * pointer returns to its starting point, but just in case, it's stopped
 * when MAXSEGS have been recorded or 2*MAXSEGS have been examined.
 */
	s.s_next = as.a_segs;
	for (i = j = k = 0; i < MAXSEGS && j < 2*MAXSEGS; j++) {
		if (s.s_next == NULL
		||  kread((KA_T)s.s_next, (char *)&s, sizeof(s)))
			break;
		if ((u_long)s.s_ops == Nl[X_SEGVN].n_value
		&&  s.s_data != NULL) {
			if (kread((KA_T)s.s_data, (char *)&sd, sizeof(sd)))
				break;
			if (sd.vn.vp) {
			
			/*
			 * This is a virtual node segment.
			 *
			 * If its vnode pointer has not been seen already,
			 * print its information.
			 */
				for (l = 0; l < k; l++) {
					if (v[l] == sd.vn.vp)
						break;
				}
				if (l >= k) {
					alloc_lfile(" txt", -1);
					process_node((caddr_t)sd.vn.vp);
					if (Lf->sf) {
						link_lfile();
						i++;
					}
					v[k++] = sd.vn.vp;
				}
			}
		}
	/*
	 * Follow the segment link to the starting point in the address
	 * space description.  (The i and j counters place an absolute
	 * limit on the loop.)
	 */
		if (s.s_next == as.a_segs)
			break;
	}
}


/*
 * readfsinfo() - read file system information
 */

static void
readfsinfo()
{
	char buf[FSTYPSZ+1];
	int i;

	if ((Fsinfomax = sysfs(GETNFSTYP)) == -1) {
		(void) fprintf(stderr, "%s: sysfs(GETNFSTYP) error: %s\n",
			Pn, strerror(errno));
		Exit(1);
	} 
	if (Fsinfomax == 0)
		return;
	if ((Fsinfo = (char **)malloc((MALLOC_S)(Fsinfomax * sizeof(char *))))
	== NULL) {
		(void) fprintf(stderr, "%s: no space for sysfs info\n", Pn);
		Exit(1);
	}
	for (i = 1; i <= Fsinfomax; i++) {
		if (sysfs(GETFSTYP, i, buf) == -1) {
			(void) fprintf(stderr,
				"%s: sysfs(GETFSTYP) error: %s\n",
				Pn, strerror(errno));
			Exit(1);
		}
		buf[FSTYPSZ] = '\0';
		if ((Fsinfo[i-1] = (char *)malloc((MALLOC_S)(strlen(buf) + 1)))
		== NULL) {
			(void) fprintf(stderr,
				"%s: no space for file system entry %s\n",
				Pn, buf);
			Exit(1);
		}
		(void) strcpy(Fsinfo[i-1], buf);
	}
}


/*
 * The ncache_addr(), ncache_load(), and ncache_lookup() functions are
 * obtained from ../common/rnch.frag.
 */
